--  Blink --  Christmas tree blink package
--  Copyright (C) 2001, 2002 Free Software Foundation, Inc.
--  Written by Stephane Carrez (stcarrez@nerim.fr)
--
--  This file is free software; you can redistribute it and/or modify it
--  under the terms of the GNU General Public License as published by the
--  Free Software Foundation; either version 2, or (at your option) any
--  later version.
--
--  In addition to the permissions in the GNU General Public License, the
--  Free Software Foundation gives you unlimited permission to link the
--  compiled version of this file with other programs, and to distribute
--  those programs without any restriction coming from the use of this
--  file.  (The General Public License restrictions do apply in other
--  respects; for example, they cover modification of the file, and
--  distribution when not linked into another program.)
--
--  This file is distributed in the hope that it will be useful, but
--  WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
--  General Public License for more details.
--
--  You should have received a copy of the GNU General Public License
--  along with this program; see the file COPYING.  If not, write to
--  the Free Software Foundation, 59 Temple Place - Suite 330,
--  Boston, MA 02111-1307, USA.

pragma No_Run_Time;

with BSP;
with BSP.IO;
with BSP.Param;
with System.Machine_Code;

package body Blink is

   use BSP;
   use BSP.IO;

   ON  : constant Light := L_1 or L_2 or L_3 or L_4  or L_5 or L_6 or L_7 or L_8;
   OFF : constant Light := 0;
   SENSE_CROSSING : constant Byte := 16#01#;

   Activity : constant String := "|/-\";
   Activity_Pos : Positive := 0;

   procedure Check_For_Reboot is
      pragma Suppress (All_Checks);
      --  Suppress checks so that there is no reference to
      --  gnat_raise_constraint_error (which is not defined).
   begin
      if (Ios.SCSR and F_RDRF) /= 0 then
         declare
            Val : Character := Ios.SCDR;
         begin
            if Val = 'R' then
               System.Machine_Code.Asm ("jmp _start");
            end if;
         end;
      end if;
      if (Ios.SCSR and F_TDRE) /= 0 then
         Ios.SCDR := Activity (Activity_Pos);
         Ios.SCCR2 := Ios.SCCR2 or F_TE;
         Activity_Pos := Activity_Pos + 1;
      end if;
   end Check_For_Reboot;

   --  Wait for N * Base_Time micro-seconds.
   procedure Sleep (N : Sleep_Delay) is
      pragma Suppress (All_Checks);
      --  Suppress checks so that there is no reference to
      --  gnat_raise_constraint_error (which is not defined).

      D : Sleep_Delay := N;
   begin
      while D /= 0 loop
         D := D - 1;

         --  Prevent gcc from removing the loop.
         System.Machine_Code.Asm ("nop", Volatile => True);
      end loop;
      Check_For_Reboot;
   end Sleep;

   --  Wait for the mains to cross 0.
   procedure Wait_Zero_Crossing is
   begin
      loop
         Check_For_Reboot;
         exit when (Ios.PORTA and SENSE_CROSSING) /= 0;
      end loop;
   end Wait_Zero_Crossing;

   --  Switch ON or OFF the lights for the specified duration.
   --  Before switching the lights, we first synchronize with
   --  the mains by waiting for it to reach 0.
   --  The lights are switched ON using PWM pulses of 250us.
   procedure Switch_Lights (Val : in Light; Duration : in Time) is
   begin
      Wait_Zero_Crossing;
      for I in 1 .. Duration loop

         --  Keep the light ON using PWM pulses on the triac gate.
         --  This loop is supposed to take 100ms.
         for J in 1 .. 5 * 20 * 2 loop
            Ios.PORTB := Val;
            Sleep (PWM_ON_Time);
            --  Turn on for 250us

            Ios.PORTB := 0;
            Sleep (PWM_OFF_Time);
            --  Turn off for 500us
         end loop;
      end loop;
   end Switch_Lights;

   --  Run the blink sequence forever.
   procedure Main_Loop is
   begin
      --  Map both I/O registers and RAM in page0
      Ios_Boot.INIT := 0;

      --  Setup SIO to receive and send at 1200 baud.
      Ios.BAUD  := F_TCLR or BSP.Param.SIO_DEF_BAUD;
      Ios.SCCR1 := 0;
      Ios.SCCR2 := F_TE or F_RE;
      loop
         for I in Speeds'Range loop
            declare
               W : Time       := Speeds (I).Wait;
               R : Repetition := Speeds (I).Repeat;
            begin
               for I in 1 .. R loop
                  Switch_Lights (ON, W);
                  Switch_Lights (OFF, W);
               end loop;
            end;
         end loop;
      end loop;
   end Main_Loop;

end Blink;
