/* panel.c -- Panel display and management interface
   Copyright 2001, 2002 Free Software Foundation, Inc.
   Written by Stephane Carrez (stcarrez@worldnet.fr)

This file is part of GEL.

GEL is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GEL is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GEL; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */


#include <sys/locks.h>
#include <sys/ports.h>
#include <string.h>
#include <assert.h>
#include <gel/event.h>
#include <gel/timer.h>
#include <gel/panel.h>
#include <gel/buttons.h>

/* Forward declarations.  */
static void panel_inactivity_timer_handler (struct timer *);
static void panel_reset_inactivity_timer (void);

static struct panel *current PAGE0_ATTRIBUTE;
static event_callout panel_button_callout;
static button_mask_t panel_prev_buttons PAGE0_ATTRIBUTE;
static struct timer panel_inactivity_timer;
static struct timer panel_refresh_timer;
static unsigned char panel_must_pop PAGE0_ATTRIBUTE;
static unsigned char panel_must_refresh PAGE0_ATTRIBUTE;
static unsigned char panel_event PAGE0_ATTRIBUTE;
static unsigned char panel_home_cursor PAGE0_ATTRIBUTE;
static struct panel visible_panel;

void
_panel_force_refresh ()
{
  panel_event = 1;
  panel_must_refresh = 1;
}

/*! Inactivity timer management.

    The inactivity timer is used to pop the current panel automatically
    when no button was pressed for some delay.  It is started when a panel
    becomes active.  It is reset each time a button is pressed.  When it
    fires, we set a global variable to indicate that the current panel
    must be popped (we can't pop it here because the handler is called
    from TCNT overflow interrupt and we may be refreshing the panel at
    the same time).

    @see panel_reset_inactivity_timer  */
static void
panel_inactivity_timer_handler (struct timer *t)
{
  panel_must_pop = 1;
  panel_event = 1;
}

/*! Setup the inactivity timer.

    If the current panel has an inactivity timer, set or reset the
    timer.

    @see panel_inactivity_timer_handler  */
static void
panel_reset_inactivity_timer (void)
{
  struct panel *p;
  
  /* Cancel the timer.  */
  timer_remove (&panel_inactivity_timer);

  /* Install it if the current panel needs it.  */
  p = current;
  assert (p != 0);

  if (p->inactivity_delay)
    {
      timer_create (&panel_inactivity_timer,
		    p->inactivity_delay,
		    panel_inactivity_timer_handler,
		    TIMER_SLOW | TIMER_UNMASK);
    }
}

static void
panel_refresh_timer_handler (struct timer *t)
{
  panel_must_refresh = 1;
  panel_event = 1;
}

static void
panel_setup_refresh_timer (unsigned long dt)
{
  timer_remove (&panel_refresh_timer);
  timer_create (&panel_refresh_timer, dt, panel_refresh_timer_handler,
		TIMER_UNMASK);
}

/*! Write the character on the panel.

    Write the character \b ch at position \b col, \b line on the panel.
    The character is not written on the LCD device but in an internal
    panel memory.  The \b panel_refresh operation synchronizes the
    panel memory and the LCD display.

    @param p    Pointer to the panel object
    @param col  Column where to write
    @param line Line where to write
    @param ch   Character to write
    @see panel_putstring, panel_refresh  */
void
panel_putchar (struct panel *p, lcd_line_t line,
	       lcd_col_t col, unsigned char ch)
{
  struct panel_line *l;

  if (p == 0)
    p = current;

  if (line >= LCD_MAX_LINES || col >= LCD_MAX_COLS)
    return;

  assert (p != 0);

  l = &p->lines[line];
  if (l->content[col] == ch)
    return;

  l->content[col] = ch;
  if (l->col_start > col)
    l->col_start = col;
  if (l->col_end < col)
    l->col_end = col;
}

/*! Write the string on the panel.

    Write the string \b str at position \b col, \b line on the panel.
    The string is not written on the LCD device but in an internal
    panel memory.  The \b panel_refresh operation synchronizes the
    panel memory and the LCD display.

    @param p    Pointer to the panel object
    @param col  Column where to write
    @param line Line where to write
    @param ch   Character to write
    @see panel_putchar, panel_refresh  */
void
panel_putstring (struct panel *p, lcd_line_t line,
		 lcd_col_t col, const char *str)
{
  unsigned char ch;

  if (p == 0)
    p = current;

  while ((ch = *str++) != 0)
    {
      if (ch == '\n')
	{
	  col = 0;
	  line++;
	  continue;
	}
      panel_putchar (p, line, col, ch);
      col++;
    }
}

void
panel_putitem (struct panel *p, lcd_line_t line, lcd_col_t col,
               enum panel_justify mode, int field_len, const char *item)
{
  const char *s;

  switch (mode)
    {
    case JUSTIFY_LEFT:
      while (field_len > 0 && *item)
        {
          panel_putchar (p, line, col, *item++);
          field_len--;
          col++;
        }
      while (field_len > 0)
        {
          panel_putchar (p, line, col, ' ');
          field_len--;
          col++;
        }
      break;

    case JUSTIFY_RIGHT:
      s = item;
      while (*s)
        s++;
      if (s > item)
        s--;

      col += field_len;
      while (field_len > 0 && s >= item)
        {
          col--;
          field_len--;
          panel_putchar (p, line, col, *s--);
        }
      while (field_len > 0)
        {
          field_len--;
          col--;
          panel_putchar (p, line, col, ' ');
        }
      break;

    default:
      assert (mode == JUSTIFY_RIGHT || mode == JUSTIFY_LEFT);
      break;
    }
}

/*! Clear the rest of the line.

    Clear the rest of the line \b line starting at \b col by
    writing spaces in it.

    @param p    Pointer to the panel object
    @param line Line to clear
    @param col  Column to start
    @see panel_putchar, panel_refresh  */
void
panel_clear_line (struct panel *p, lcd_line_t line, lcd_col_t col)
{
  if (p == 0)
    p = current;

  assert (p != 0);
  for (; col < LCD_MAX_COLS; col++)
    panel_putchar (p, line, col, ' ');
}

/*! Return the character displayed on the panel.

    @param p   Null or pointer to the panel object
    @param col Column number
    @param line Line number
    @return The character at the given position
    @see panel_putchar  */
unsigned char
panel_getchar_at (struct panel *p, lcd_line_t line, lcd_col_t col)
{
  if (p == 0)
    p = current;

  if (col >= LCD_MAX_COLS || line >= LCD_MAX_LINES)
    return 0;

  return p->lines[line].content[col];
}

/* Synchronize the panel line with the LCD line.  */
static void
panel_refresh_line (struct panel_line *line, struct panel_line *vline,
		    lcd_line_t line_num)
{
  lcd_col_t i;

  if (line->col_start > line->col_end)
    return;

  /* Only update the part of the line which has changed.  */
  lcd_goto (line->col_start, line_num);
  for (i = line->col_start; i <= line->col_end; i++)
    {
      vline->content[i] = line->content[i];
      lcd_putchar (line->content[i]);
    }
  line->col_start = LCD_MAX_COLS + 1;
  line->col_end = 0;
}

/*! Refresh the panel.

    Refresh the LCD display to match the content with the active panel.
    This function optimizes by only updating the LCD part which has
    changed. This is the only function which calls the LCD print operations.

    @see panel_putchar, panel_putstring, panel_redraw  */
void
panel_refresh ()
{
  lcd_line_t i;
  struct panel *p = current;
  int bstate;

  /* If buttons are enabled, they must be deactivated so that we
     can send LCD commands.  Do this only if there is really something
     to refresh.  */
  bstate = button_is_enabled ();
  if (bstate)
    {
      if (panel_need_refresh () == 0)
	return;

      button_disable ();
    }

  assert (p != 0);

  if (panel_home_cursor)
    {
      panel_home_cursor = 0;
      lcd_home ();
    }

  /* Refresh the LCD screen.  */
  lcd_cursor (p->cursor_visible);
  for (i = 0; i < LCD_MAX_LINES; i++)
    panel_refresh_line (&p->lines[i], &visible_panel.lines[i], i);

  /* Setup the cursor for the panel.  */
  if (p->cursor_visible)
    {
      lcd_goto (p->cursor_col, p->cursor_line);
    }
  visible_panel.cursor_visible = p->cursor_visible;

  /* Restore the buttons.  */
  if (bstate)
    button_enable ();
}

/*! Touch the panel.

    Forget any optimization about the LCD panel display.

    @see panel_refresh  */
void
panel_touch ()
{
  lcd_line_t i;
  struct panel *p = current;

  assert (p != 0);
  for (i = 0; i < LCD_MAX_LINES; i++)
    {
      p->lines[i].col_start = 0;
      p->lines[i].col_end = LCD_MAX_COLS - 1;
    }
}

/*! Returns true if the panel must be refreshed.

    Returns true if the LCD display is not synchronized with the
    active panel and a call to \b panel_refresh is necessary.

    @return true if the panel must be refreshed.
    @see panel_refresh, panel_touch  */
int
panel_need_refresh ()
{
  struct panel *p = current;
  struct panel_line *line;
  lcd_line_t i;

  assert (p != 0);

  /* If the cursor changed, the panel must be refreshed.  */
  if (p->cursor_visible != visible_panel.cursor_visible)
    return 1;

  for (i = 0; i < LCD_MAX_LINES; i++)
    {
      line = &p->lines[i];
      if (line->col_start <= line->col_end)
	{
	  struct panel_line *vline = &visible_panel.lines[i];
	  lcd_col_t j;

	  for (j = line->col_start; j <= line->col_end; j++)
	    {
	      if (vline->content[j] != line->content[j])
		return 1;
	    }
	}
    }

  return 0;
}

/*! Panel is visible.

    Return whether the panel is visible.

    @param p   Pointer to the panel object
    @return  true if the panel is visible
    @see panel_push  */
int
panel_is_visible (struct panel *p)
{
  return p == current || p->next != 0;
}

static void
panel_install (struct panel *p)
{
  panel_must_pop = 0;
  panel_must_refresh = 1;
  panel_home_cursor = 1;
  current = p;

  panel_reset_inactivity_timer ();
  if (p->to_enter)
    p->to_enter (p);

  visible_panel.cursor_visible = 2;
  panel_touch ();
}

/*! Push the panel.

    Push the panel \b p making it the active panel.  The previous
    active panel is kept for later re-activation with panel_pop.

    @param p   Pointer to the panel object
    @see   panel_pop  */
void
panel_push (struct panel *p)
{
  if (panel_is_visible (p))
    return;

  if (current && current->to_exit)
    current->to_exit (current);

  p->next = current;
  panel_install (p);
}

/*! Pop the active panel moving to the previous one.

    Pop the active panel and redraw the previous one.
    The to_exit handler of the active panel is called before deactivating
    it.  The to_enter handler of the previous panel is called
    to activate it.

    @see panel_push, panel_refresh  */
void
panel_pop (void)
{
  struct panel *p;

  p = current;
  if (p == 0 || p->next == 0)
    return;

  if (p->to_exit)
    p->to_exit (p);

  panel_install (p->next);
  p->next = 0;
}

/* Button event handler.

   Get the buttons and detect those which are released and pressed.
   Call first the panel \b to_input callback for each button which
   is released.  Then, call the panel \b to_input callback for each
   button pressed.

*/
static void
panel_button_event_handler (event_callout *callout, event *ev)
{
  button_mask_t mask;
  button_mask_t delta;
  panel_input_t input;

  mask = button_get_pressed ();
  delta = mask ^ panel_prev_buttons;

  assert (current != 0);
  input = current->to_input;

  assert (input != 0);

  /* We got something, reset the inactivity timer.  */
  panel_reset_inactivity_timer ();

  /* Send a KEY_RELEASE note for each button which was released.  */
  if (delta & panel_prev_buttons)
    {
      unsigned i;

      for (i = 0; i <= GEL_MAX_BUTTONS; i++)
        {
          if (delta & (1 << i) & panel_prev_buttons)
            input (current, KEY_RELEASE, i);
        }
    }

  /* Send a KEY_PRESS note for each button which was pressed.  */
  if (delta & mask)
    {
      unsigned i;

      for (i = 0; i <= GEL_MAX_BUTTONS; i++)
        {
          if (delta & (1 << i) & mask)
            input (current, KEY_PRESS, i);
        }
    }
  panel_prev_buttons = mask;

  panel_refresh ();
}

void
panel_setcursor (struct panel *p, lcd_line_t line, lcd_col_t col)
{
  if (p == 0)
    p = current;

  assert (p != 0);

  p->cursor_col = col;
  p->cursor_line = line;
  visible_panel.cursor_visible = 2;
}

void
panel_show_cursor (struct panel *p, int mode)
{
  if (p == 0)
    p = current;

  assert (p != 0);

  p->cursor_visible = mode;
  visible_panel.cursor_visible = 2;
}

/*! Panel main loop.

    The panel main loop is the infinite loop which waits for buttons
    to be pressed or released and activates the input and refresh of
    the active panel.

    @see panel_push, panel_pop  */
void __attribute__ ((noreturn))
panel_loop ()
{
  assert (current != 0);
  while (1)
    {
      unsigned short s;

      /* Refresh the panel before waiting for buttons.  */
      if (panel_must_refresh)
	{
	  unsigned long dt;

	  panel_must_refresh = 0;
	  if (current->to_refresh)
	    {
	      dt = current->to_refresh (current);
	      panel_setup_refresh_timer (dt);

	      /* Update the LCD.  */
	      panel_refresh ();
	    }
	}
      
      /* Wait for 31 TCNT overflow and process the event callbacks which
	 are fired.  */
      s = lock ();
      if (panel_event == 0)
	{
	  restore (s);
	  event_loop (31L * 65536L, &panel_event);
	}
      else
	{
	  panel_event = 0;
	  restore (s);
	}

      /* If the inactivity timer fired, the user is inactive for too much
	 time and we must pop the active panel.  */
      if (panel_must_pop)
        {
          panel_pop ();
        }
    }
}

/*! Create the panel object.

    @param p    Pointer to the panel object.  */
void
panel_create (struct panel *p)
{
  lcd_line_t i;

  p->visible = 0;
  p->cursor_visible = 0;
  p->cursor_col = 0;
  p->cursor_line = 0;
  p->next = 0;

  for (i = 0; i < LCD_MAX_LINES; i++)
    {
      struct panel_line *line;

      line = &p->lines[i];
      line->col_start = LCD_MAX_COLS + 1;
      line->col_end = 0;
      memset (line->content, ' ', LCD_MAX_COLS);
    }
}

/*! Initialize the panel module.

    @see panel_create  */
void
panel_initialize (void)
{
  event_register (&button_change_event, &panel_button_callout,
                  panel_button_event_handler, 0);
  panel_create (&visible_panel);
}

